import uuid
import secrets
from typing import Optional
from sqlmodel import Field, SQLModel, select

class Users(SQLModel, table = True):    
    __tablename__ = 'users'

    id: int = Field(unique = True, primary_key = True)
    uuid: str = Field(unique = True)
    user_name: str = Field(unique = True)
    email: str = Field(unique = True)
    password: str = Field(unique = False)
    status: int
    date_login: Optional[int] = None
    date_created: Optional[int] = None
    date_deleted: Optional[int] = None

    @staticmethod
    def find_user_or_create(session, email, user_name = None):
        user = Users.find_by_email(session,email)

        if not user:
            # create
            user = Users(
                uuid = str(uuid.uuid4()),
                email = email,
                user_name = user_name,
                status = 1,
                password = secrets.token_hex(32)  
            )
            session.add(user)
            session.commit()
            
            return Users.find_by_email(email)
        else:
            return user
      
    
    @staticmethod
    def find_by_email(session, email):
        stm = select(Users).where(Users.email == email)
        return session.exec(stm).first()
    

    @staticmethod
    def find_user(session,id):
        stm = (
            select(Users)
            .where(
                (Users.id == id) |
                (Users.uuid == id)
            )
            .where(Users.status == 1)
            .where(Users.date_deleted == None)
        )

        return session.exec(stm).first()
    


from typing import Optional
from sqlmodel import Field, SQLModel, select

class Options(SQLModel, table = True):     
    __tablename__ = 'options'

    id: int = Field(unique = True, primary_key = True)
    key: str
    value: str | None
  
    @staticmethod
    def save_option(session, key, value):
        stm = (
            select(Options.value).where(Options.key == key)
        )
        model = session.exec(stm).first()
        
        if model == None:
            option = Options(
                key = key,
                value = value
            )
            session.add(option)
            session.commit()
        else:
            model.value = value
            return model.save()
            

    @staticmethod
    def get_option(session, key, default = None):
        stm = select(Options.value).where(Options.key == key)
        model = session.exec(stm).first()
        
        if not model:
            return default
        
        return model.value


    @staticmethod
    def increment(session, key, value = 1):

        model, created = Options.get_or_create(
            key = key,
            defaults = {
                'value': 1
            }
        )

        if not created:
            return (Options.update(value = Options.value + 1)
                .where(Options.key == value)
                .execute())
        


from arikaim.core.controller import Controller, get
from arikaim.core.services import services

class ServicesList(Controller):

    @get
    async def get(self, request, data):  
        self.field('services',services.services)
        self.message('Service list')
        


from typing import Union
from arikaim.core.utils import create_action
from arikaim.core.logger import logger
import traceback

class RunAction():

    def run_action(self, 
        class_name, 
        service_name: Union[str,None] = None, 
        module_name: Union[str,None] = None, 
        options: dict = {}
    ):
        try:
            if service_name and module_name:
                action = create_action(service_name,module_name,class_name,options)
            else:
                action = class_name(options)

            action.run()

            if action.has_error() == True:              
                self.error(action.error)
                return 
        
            for key, value in action.results.items():               
                self.field(key,value)
            
            self.field('message',action.get('message','Success!'))
            
        except Exception as error:      
            logger.error(traceback.format_exc())    
            self.error('Error run action: ' + str(error))


from rich import print
from rich.prompt import Prompt
import click
from arikaim.core.app import app
from arikaim.core.services import services
from arikaim.core.packages import save_package_descriptor,load_package_descriptor


@click.group(name = 'services')
def services_group():
    pass

@click.command(name = 'enable')
def enable_service():
    print('Enable service')
    print('')
    service_name = get_service_name()
    if service_name == False:
        return
    
    descriptor = load_package_descriptor(service_name)
    descriptor['disabled'] = False
    save_package_descriptor(service_name,descriptor)
    print('[green]Done')
    

@click.command(name = 'disable')
def disable_service():
    print('Disable service')
    print('')
    service_name = get_service_name()
    if service_name == False:
        return
    
    descriptor = load_package_descriptor(service_name)
    descriptor['disabled'] = True
    save_package_descriptor(service_name,descriptor)
    print('[green]Done')
    
def get_service_name():
    service_name = Prompt.ask("Service name")
    if not service_name:
        print('[red]Service name is required')
        return False
    
    services.scan_services()
    if service_name not in services:
        print('[red]Service not exist')
        return False

    return service_name

@click.command(name = 'list')
def list_services():
    services.scan_services()
    print('')
    print('Services list')
   
    for name, item in services.services.items():
        print('')
        print(item["title"], end = " ")
        if item['disabled'] == True:
            print('[red] disabled',end = " ")
        else:
            print('[green] enabled',end = " ")
    
    print('')
    print('')


services_group.add_command(list_services)
services_group.add_command(enable_service)
services_group.add_command(disable_service)


import html
from schema import *
from starlette.endpoints import HTTPEndpoint
from starlette.types import Receive, Scope, Send
from starlette.responses import Response

from arikaim.core.collection.descriptor import PropertiesDescriptor
from arikaim.core.response import ApiResponse

class Controller(HTTPEndpoint):
  
    def __init__(self, scope: Scope, receive: Receive, send: Send):
        """Constructor. """
        super().__init__(scope, receive, send)
        # response fields
        self._status = 'ok'
        self._errors = []
        self._code = 200 
        self._fields = {}
        self._descriptor = None
        self._schema = None
        self._media_type = 'application/json'
        self.boot()
        
    def media_type(self, value):
        self._media_type = value

    def schema(self, schema_def):
        self._schema = Schema(schema_def)

    def validate(self, data):
        if self._schema == None:
            return True
    
        try:
            return self._schema.validate(data)
        except SchemaError as e:
            self.error(str(e))
            return False

    def boot(self):
        pass

    def response(self):       
        if self._media_type == 'application/json':
            return ApiResponse(
                self.get_result(), 
                media_type = self._media_type
            )
        else:
            return Response(
                content = self._fields['content'], 
                media_type = self._media_type
            )
        
    def message(self, value):
        self._fields['message'] = value

    def field(self, key, value):
        self._fields[key] = value

    def content(self, value):
        self._fields['content'] = value

    def error(self, error):
        self._errors.append(error)

    @property
    def status(self,value):
        self._status = value
    
    @property
    def code(self, value):
        self._code = value

    def get_result(self):  

        if (len(self._errors) > 0):
            self._status = 'error'

        return {
            'result': self._fields,
            'status': self._status,
            'errors': self._errors,
            'code'  : self._code
        }

    @classmethod
    def descriptor(cls):
        descriptor = PropertiesDescriptor()
        cls.init_descriptor(descriptor)
   
        return descriptor
    
    @classmethod
    def init_descriptor(cls, descriptor):               
        pass


async def get_request_data(request):
    query_params = {}
    path_params = {}

    if request.query_params:
        query_params = request.query_params
    if request.path_params:
        path_params = request.path_params
  
    try:
        if request.headers['content-type'] == 'application/json':
            body = await request.json()
            data = {**query_params,**path_params,**body}
            # sanitize
            for key, val in body.items():
                if type(val) == str:
                    data[key] = html.escape(val)

        else:
            form = await request.form()
            data = dict(form)
    except: 
        return {**query_params,**path_params}
   
    return data

def decorator(func):
    async def wrap(self, request):
        data = await get_request_data(request)
        if self.validate(data) == False:
            return self.response()
        else:
            await func(self,request,data)
        return self.response()
    
    return wrap

# Decorators
def put(func):
    return decorator(func)

def post(func):
    return decorator(func)

def get(func):
    return decorator(func)

def delete(func):
    return decorator(func)


from starlette.authentication import SimpleUser
from arikaim.core.db.models.users import Users
from arikaim.core.db.models.access_tokens import AccessTokens
from arikaim.core.redis import redis
from arikaim.core.utils import php_unserialize
from arikaim.core.access.user import AuthUser

class RedisPHPSessionAuthProvider:
    
    def __init__(self):
        pass

    def authenticate(self, credentails):
        session_id = credentails.cookies.get('PHPSESSID')       
        if not session_id:
            return False
        
        try:
            session_data = redis.get(session_id)
            if not session_data:
                return False
            
            session = php_unserialize(str(session_data))
          
            if 'auth.id' in session:
                auth_id = session['auth.id']
            else:
                return False
                 
            user = Users.find_user(auth_id)
            if user == None:
                return False
            else:
                return AuthUser(id = user.id, uuid = user.uuid, username = user.user_name, email = user.email)
        
        except (Users.DoesNotExist, AccessTokens.DoesNotExist):
            return False




class Job:

    def __init__(self, name = None, interval = None, datetime = None):     
        self._name = name
        self._interval = interval
        self._datetime = datetime

    def __call__(self):
        self.execute()

    def execute(self):
        raise Exception('Job execute methos not implemented')
    
    @property
    def interval(self):
        return self._interval
    
    @property
    def datetime(self):
        return self._datetime


from typing import Optional
from sqlmodel import Field, SQLModel

class AccessTokens(SQLModel, table = True):
    __tablename__ = 'access_tokens'

    id: int = Field(unique = True, primary_key = True)
    uuid: str = Field(unique = True)
    token: str = Field(unique = True)
    status: int
    type: int
    user_id: int
    date_created: Optional[int] = None
    date_expired: Optional[int] = None
 


import click
import sys
import os.path
from rich import print, pretty

from arikaim.core.server import arikaim_server
from arikaim.core.services import services
from arikaim.core.app import app
from arikaim.core.console.config import config
from arikaim.core.console.install import install
from arikaim.core.console.packages import packages
from arikaim.core.console.services import services_group
from arikaim.core.console.queue import queue_group
from arikaim.core.logger import logger
from arikaim.core.path import Path
from arikaim.core.utils import load_module

@click.group(invoke_without_command = True)
@click.pass_context
def main(ctx):
    pretty.install()

    print("")
    print("[blue]Arikaim Services Server")
    print("[green]version [white]" + arikaim_server.version)
    print("")

    if not ctx.invoked_subcommand:
        click.echo(ctx.get_help())

@click.command()
@click.argument('mode', required = False)
@click.option('--path', required = False)
def run(mode = None, path = None):
    # run server
    if mode == 'dev':
        mode = True
 
    arikaim_server.run(mode,path)

# load service console commands
def load_services_commands():
    app.system_init()
    services.scan_services()

    for service in services.services:
        logger.info('Service: ' + service + ' load console commands ...')
        app.load_console_commands(service)

@click.command()
def info():
    logger.info('Base Bath ' + Path.base(False)) 
    logger.info('Python Version ' + sys.version) 
   
@click.command(name = 'console')
@click.argument('name', required = True)
def console_cmd(name: str):
    print("[blue]Console")
    print("")

    app.system_init()
    services.scan_services()

    path = os.path.join(Path.console(name),'console.py')
    if os.path.isfile(path) == False:
        print('[red]Console app not exist for this service')
        print("")
        return
    
    load_module('console',path)

main.add_command(run)
main.add_command(services_group)
main.add_command(packages)
main.add_command(config)
main.add_command(install)
main.add_command(info)
main.add_command(queue_group)
main.add_command(console_cmd)


import abc
from arikaim.core.collection.descriptor import PropertiesDescriptor

class Action:

    def __init__(self, options: dict = {}):
        self._name = ''
        self._title = ''
        self._description = ''
        self._error = None
        self._descriptor = None
        
        self._options = options
        self._result = {}
        self.init()

    @property
    def name(self):
        return self._name
    
    @name.setter
    def name(self, name: str):
        self._name = name

    @property
    def title(self):
        return self._title
    
    @title.setter
    def title(self, title: str):
        self._title = title

    @property
    def description(self):
        return self._description
    
    @description.setter
    def description(self, description: str):
        self._description = description
    
    def __call__(self):
        return self.run()
    
    def init(self):
        pass

    @abc.abstractmethod
    def run(self):
        """Implement action run method"""
        return

    def has_error(self):
        if not self._error:
            return False
        else:
            return True

    def set_error(self, error: str):
        self._error = error
        return self
    
    @property
    def error(self):
        return self._error
    
    def options(self, options: dict):
        self._options = options
        return self
    
    def option(self, name: str, value: any):
        self._options[name] = value
        return self
    
    def get_option(self, name: str, default: any = None):
        if name in self._options:
            return self._options[name]
        
        return default

    def get_options(self):
        return self._options

    def result(self, name: str, value: any):
        self._result[name] = value
        return self
    
    @property
    def results(self):
        return self._result
    
    def get(self, name: str, default: any = None):
        if name in self._result:
            return self._result[name]
        
        return default
    
    @classmethod
    def descriptor(cls):
        descriptor = PropertiesDescriptor()
        cls.init_descriptor(descriptor)
   
        return descriptor
    
    @classmethod
    def init_descriptor(cls, descriptor):               
        pass

    


from arikaim.core.controller import Controller, get
from arikaim.core.services import services

class ServiceRoute(Controller):

    @get
    async def get(self, request, data):       
        service = services.get_service(data['name'])
        route_name = data['route_name']
        result = {} 
        
        for route in service.routes.routes:
           
            if route.name == route_name:
                descriptor = route.endpoint.descriptor()
                result = {
                    'title': descriptor.title,
                    'description': descriptor.description,
                    'path': service.mount_path + route.path,                         
                    'name': route.name,
                    'methods': list(route.methods),
                    'params': descriptor.params,
                    'result': descriptor.result_fields,
                    'example_values': descriptor.get_example_values()     
                }
                break

        self.field('name',data['name'])
        self.field('path',service.mount_path + route.path)       
        self.field('route',result)
        self.message('Service route')
    


import gc

from starlette.middleware.base import BaseHTTPMiddleware
from arikaim.core.utils import get_process_memory
from arikaim.core.logger import logger

class SystemMiddleware(BaseHTTPMiddleware):

    async def dispatch(self, request, call_next):
        
        gc.collect()
        logger.info('Used memory ' + str(get_process_memory()))

        response = await call_next(request)
        return response



import os

class Path:
    base_path = ''

    @staticmethod
    def base(relative = True):
        if Path.base_path is None:
            Path.base_path = os.getcwd()

        if relative == True:
            return Path.base_path
           
        return os.path.abspath(Path.base_path)

    @staticmethod
    def arikaim(relative = True):
        return os.path.join(Path.base(relative),'arikaim')
    
    @staticmethod
    def view(path: str = '', relative = True):       
        return os.path.join(Path.arikaim(relative),'view',path.strip('/'))
    
    @staticmethod
    def templates(relative = True):       
        return os.path.join(Path.view(relative = relative),'templates')
    
    @staticmethod
    def template(name: str, relative: bool = True):       
        return os.path.join(Path.templates(relative),name)
    
    @staticmethod
    def model(service_name, relative: bool = True):
        if not service_name:
            return os.path.join(Path.base(relative),'db','models')
        else:
            return os.path.join(Path.services(service_name, relative = relative),'models')

    @staticmethod
    def actions(service_name, relative: bool = True):
        if not service_name:
            return os.path.join(Path.base(relative),'actions')
        else:
            return os.path.join(Path.services(service_name, relative = relative),'actions')
        
    @staticmethod
    def job(service_name, relative: bool = True):
        if not service_name:
            return os.path.join(Path.base(relative),'queue','jobs')
        else:
            return os.path.join(Path.services(service_name, relative = relative),'jobs')

    @staticmethod
    def console(service_name, relative = True):
        return os.path.join(Path.services(service_name, relative = relative),'console')

    @staticmethod
    def storage(path: str = '', relative = True):       
        return os.path.join(Path.base(relative),'arikaim','storage',path.strip('/'))
        
    @staticmethod
    def services(name = '', relative = True):
        services_path = os.path.join(Path.base(relative),'arikaim','services')
        if name != "":
            return os.path.join(services_path,name)
        return services_path

    @staticmethod
    def config(file_name = None, relative = True):
        path = os.path.join(Path.base(relative),'arikaim','config')
        if (file_name != None): 
            path = os.path.join(path,file_name) 
        return path
        


import redis as r

redis = None

def redis_connect(config = None):
    global redis
    if config == False:
        redis = None
        return None
    
    if not config:
        redis = r.Redis()
    else:
        redis = r.Redis(config)
     
    return redis



from arikaim.core.db.models.users import Users
from arikaim.core.db.models.access_tokens import AccessTokens
from arikaim.core.access.user import AuthUser
from arikaim.core.db.db import db
from sqlmodel import Session, select
from sqlalchemy.exc import SQLAlchemyError

class TokenAuthProvider:
    
    def authenticate(self, credentails):

        if "Authorization" not in credentails.headers:
            return False
        try:
            token = credentails.headers["Authorization"]
      
            with Session(db.engine) as session:
                statement = select(AccessTokens).where(AccessTokens.token == token)
                access_tokens = session.exec(statement).first()
                if not access_tokens:
                    session.close()
                    return False
                
                statement = select(Users).where(Users.id == access_tokens.user_id)
                user = session.exec(statement).first()
                session.close()
                
                return AuthUser(id = user.id, uuid = user.uuid, username = user.user_name, email = user.email)
        except Exception as e:
            print(e)
            return False


from typing import Optional
from sqlmodel import Field, SQLModel

class Permissions(SQLModel, table = True):
    __tablename__ = 'permissions'

    id: int = Field(unique = True, primary_key = True)
    uuid: str = Field(unique = True)
   
    name: str = Field(unique = True)
    slug: str = Field(unique = True)
    title: str
    extension_name: str
    description: str
    deny: int



from typing import Optional
from sqlmodel import Field, SQLModel

class PermissionRelations(SQLModel, table = True):
    __tablename__ = 'permission_relations'

    id: int = Field(unique = True, primary_key = True)
    uuid: str = Field(unique = True)
    read: int
    write: int
    delete: int
    execute: int
    permission_id: int
    relation_id: int
    relation_type: str



from arikaim.core.response import ApiResponse
from starlette.middleware.base import BaseHTTPMiddleware
from arikaim.core.logger import logger
import traceback


class ExceptionHandlerMiddleware(BaseHTTPMiddleware):
    
    async def dispatch(self, request, call_next):
        try:
            return await call_next(request)
        except Exception as e:                 
            logger.error(str(e))
          
            return ApiResponse({
                'result': None,
                'status': 'error',
                'errors': ['Server error'],
                'code'  : 500
            })  



from typing import Optional
from sqlmodel import Field, SQLModel, select

class Drivers(SQLModel, table = True):
    __tablename__ = 'drivers'

    id: int = Field(unique = True, primary_key = True)
    uuid: str = Field(unique = True)
    name: str = Field(unique = True)
    title: str
    description: str
    version: str
    config: str
   
    @staticmethod
    def get_config(session, name):
        stm = select(Drivers).where(Drivers.name == name)
        model = session.exec(stm).first()
        
        if not model:
            return None
        else:    
            return model.config



from starlette.authentication import AuthenticationBackend, AuthCredentials
from arikaim.core.access.auth_error import AuthError
from arikaim.core.access.access import access
from arikaim.core.access.user import AuthUser

class AuthMiddleware(AuthenticationBackend):
    
    def __init__(self, auth_providers = []):
        self._providers = auth_providers

    def add_provider(self, auth_provider):
        if not auth_provider in self._providers:
            self._providers.append(auth_provider)

    async def authenticate(self, conn): 
        try:
            for provider in self._providers:
                user = access.authenticate(conn,provider)
            
                if isinstance(user,AuthUser) == True:
                    return AuthCredentials(["authenticated"]), user
        except:
            raise AuthError(status_code = 401)
        
        raise AuthError(status_code = 401)



from starlette.authentication import BaseUser

class AuthUser(BaseUser):

    def __init__(self, username: str = '', id = None, email: str = '', uuid: str = '', authenticated: bool = True) -> None:
        self._username = username
        self._id = id
        self._email = email
        self._uuid = uuid
        self._authenticated = authenticated

    @property
    def is_authenticated(self) -> bool:
        return self._authenticated

    @property
    def id(self):
        return self._id
    
    @property
    def uuid(self) -> str:
        return self._id
    
    @property
    def email(self) -> str:
        return self._email
    
    @property
    def display_name(self) -> str:
        return self._username



import json
from arikaim.core.db.db import load_model_class
from arikaim.core.collection.properties import Properties

class DriverManager():
    _instance = None

    def __init__(self):     
        pass

    def __new__(cls, *args, **kwargs):
        if not isinstance(cls._instance, cls):
            cls._instance = object.__new__(cls, *args, **kwargs)
        return cls._instance
    
    def get_config(self, session, driver_name: str):
        Drivers = load_model_class('Drivers','drivers')
        json_text = Drivers.get_config(session,driver_name)
        if not json_text:
            return None
        
        data = json.loads(json_text)
        properties = Properties(data)
 
        return properties


drivers = DriverManager()



from jinja2 import Environment, FileSystemLoader, select_autoescape
from arikaim.core.path import Path

class View():
    _instance = None

    def __init__(self):   
        self._templates_path = Path.templates()

        self._env = Environment(
            loader = FileSystemLoader(self._templates_path),
            autoescape = select_autoescape()
        )  
        pass

    def __new__(cls, *args, **kwargs):
        if not isinstance(cls._instance, cls):
            cls._instance = object.__new__(cls, *args, **kwargs)
        return cls._instance


    def render_template(self, path, data = {}):
        template = self._env.get_template(path)
        content = template.render(data)
        return content



view = View()



from os.path import exists
import subprocess
import json

from arikaim.core.db.models.users import Users
from arikaim.core.db.models.access_tokens import AccessTokens
from arikaim.core.utils import php_unserialize
from arikaim.core.logger import logger
from arikaim.core.path import Path 
from arikaim.core.access.user import AuthUser

class PhpSessionAuthProvider:
    
    def __init__(self):
        self._php_session_path = None
        self.resolve_session_path()

    def authenticate(self, credentails):
        session_id = credentails.cookies.get('PHPSESSID')
        if not session_id:
            return False
        try:
            session_data = self.read_php_session(session_id)
            if 'auth.id' not in session_data:
                return False
  
            user = Users.get(Users.id == session_data['auth.id'])

            return AuthUser(id = user.id, uuid = user.uuid, username = user.user_name, email = user.email)
        except (Users.DoesNotExist, AccessTokens.DoesNotExist):
            return False

    def read_php_session(self, session_id):
        file_name = self._php_session_path + '/sess_' + session_id

        if exists(file_name) == False:
            return {}
      
        logger.info('Read php session file: ' + file_name)
        try:
            file = open(file_name,'r')
            content = file.read()
            file.close()
        except IOError:
            return False
        
        return php_unserialize(content)
    
    def resolve_session_path(self):
        try:
            response = subprocess.Popen(
                'php cli session:info --output json',
                shell = True,
                cwd = Path.base(),
                stdout = subprocess.PIPE          
            )

            json_text = response.stdout.read()
            data = json.loads(json_text)
        except BaseException:
            self._php_session_path = None
            return
        
        self._php_session_path = data['save_path']


import os
from arikaim.core.path import Path


def create_storage_path(path: str, mode = 0o666):
    storage_path = Path.storage(path)
    
    if os.path.isdir(storage_path) == False:
        os.mkdir(storage_path,mode) 
        return os.path.isdir(storage_path)
    else:
        return True


from arikaim.core.controller import Controller, get
from arikaim.core.services import services

class ServiceRoutes(Controller):

    @get
    async def get(self, request, data):         
        service = services.get_service(data['name'])
        
        routes = []
        if service != None:
            routes = self.get_routes(service)
       
        self.field('name',data['name'])
        self.field('routes',routes)
        self.message('Service routes')
       
    def get_routes(self,service):
        routes = []
        for route in service.routes.routes:
            descriptor = route.endpoint.descriptor()
            routes.append({
                'title': descriptor.title,
                'description': descriptor.description,               
                'path': service.mount_path + route.path,
                'name': route.name,                
                'methods': list(route.methods),
                'params': descriptor.params,
                'result': descriptor.result_fields          
            })

        return routes


from starlette.exceptions import HTTPException
from starlette.requests import Request
from arikaim.core.response import ApiResponse

async def server_error(request: Request, exc: HTTPException): 
    return ApiResponse({
        'result': '',
        'details': exc.detail,
        'status': exc.status_code,
        'errors': ['Server errror'],
        'code'  : exc.status_code
    }) 

async def unauthorized_error(request: Request, exc: HTTPException): 
    return ApiResponse({
        'result': '',
        'details': exc.detail,
        'status': exc.status_code,
        'errors': ['Access Denied'],
        'code'  : exc.status_code
    })  

async def not_found_error(request: Request, exc: HTTPException): 
    return ApiResponse({
        'result': '',
        'details': exc.detail,
        'status': exc.status_code,
        'errors': ['Not found'],
        'code'  : exc.status_code
    })  

async def api_call_quota_limit_error(request: Request, exc: HTTPException): 
    return ApiResponse({
        'result': '',
        'details': exc.detail,
        'status': exc.status_code,
        'errors': ['Api call quota limit reached error'],
        'code'  : exc.status_code
    }) 

async def api_call_quota_billing_error(request: Request, exc: HTTPException): 
    return ApiResponse({
        'result': '',
        'details': exc.detail,
        'status': exc.status_code,
        'errors': ['Api call billing limit reached'],
        'code'  : exc.status_code
    }) 

async def http_exception(request: Request, exc: HTTPException):
    return ApiResponse({
        'result': '',
        'details': exc.detail,
        'status': exc.status_code,
        'errors': ['Http error'],
        'code'  : exc.status_code
    })

async def handle_error(request: Request, exc: HTTPException):
    return ApiResponse({
        'result': '',
        'details': exc.detail,
        'status': exc.status_code,
        'errors': ['Error '],
        'code'  : exc.status_code
    })

error_handlers = {
    401: unauthorized_error,
    404: not_found_error,
    500: server_error,
    429: api_call_quota_limit_error,
    430: api_call_quota_billing_error,
    HTTPException: http_exception,
    Exception: handle_error
}



from arikaim.core.utils import *
from starlette.routing import Route, Mount, WebSocketRoute
from starlette.middleware import Middleware
from starlette.middleware.authentication import AuthenticationMiddleware
from arikaim.core.access.middleware import AuthMiddleware
from arikaim.core.path import Path

class Service:
    
    def __init__(self, name: str, mount_path = None):
        self._routes = []      
        self._middlewares = []     
        self._route_middlewares = []     
        self._name = name
        self._mounth_path = mount_path
    
    def boot(self):
        pass

    def init_routes(self):
        pass

    def init_container(self):
        pass

    def add_route(self, method, path, endpoint, name = None, auth_providers = None):
        middleware = None
        if auth_providers != None:
            backend = AuthMiddleware(auth_providers)
            middleware = [Middleware(AuthenticationMiddleware, backend = backend)]
            
        self._routes.append(Route(
            path,
            endpoint = endpoint,
            name = name,
            methods = method,
            middleware = middleware
        ))           

    def add_websocket_route(self, path, endpoint):       
        self._routes.append(WebSocketRoute(path,endpoint = endpoint))

    def add_middleware(self, middleware_class):
        self._middlewares.append(Middleware(middleware_class))     

    def add_auth_middleware(self, auth_providers):
        backend = AuthMiddleware(auth_providers)
        self._middlewares.append(Middleware(AuthenticationMiddleware, backend = backend))  

    def path(self, path):
        self._mounth_path = path

    @property
    def mount_path(self): 
        if self._mounth_path == None:
            return str('/' + self._name)
        else:
            return str('/' + self._mounth_path)
        
    @property
    def routes(self):   
        if len(self._routes) == 0:
            return None
        else:
            return Mount(
                self.mount_path, 
                routes = self._routes, 
                name = self._name, 
                middleware = self._middlewares
            )
        

def create_instace(service_name):
    service_class = load_class(Path.services(service_name),service_name,service_name.capitalize())
    return service_class(service_name) 


import click
from rich import print
from arikaim.core.path import Path
from arikaim.core.console.templates.config_file import content
from arikaim.core.app import app
from arikaim.core.queue.queue import queue
from arikaim.core.services import services

@click.group(name = 'queue')
def queue_group():
    pass

@click.command()
def run():
    app.system_init()
    services.scan_services()
   
    queue.run()

@click.command()
def jobs():
   
    print(queue.jobs)

queue_group.add_command(run)
queue_group.add_command(jobs)


from arikaim.core.service import Service
from arikaim.core.admin.controllers.services_list import ServicesList
from arikaim.core.admin.controllers.service import ServiceDescriptor
from arikaim.core.admin.controllers.service_routes import ServiceRoutes
from arikaim.core.admin.controllers.service_route import ServiceRoute

class AdminService(Service):

    def init_routes(self):
        self.path('help')
        self.add_route(["GET"],"/services",ServicesList)
        self.add_route(["GET"],"/service/details/{name}",ServiceDescriptor)
        self.add_route(["GET"],"/service/routes/{name}",ServiceRoutes)
        self.add_route(["GET"],"/service/route/{name}/{route_name}",ServiceRoute)


from starlette.middleware.base import BaseHTTPMiddleware
import bleach
import json

class SanitizeMiddleware(BaseHTTPMiddleware):

    def __init__(self, app):
        super().__init__(app)

    async def dispatch(self, request, call_next):
        await self.set_body(request)

        if request.method == 'POST' or request.method == 'PUT':
            json_body = await request.json()
            sanitize_body = {key: self.sanitize_array(value) if isinstance(
                value, list) else bleach.clean(value) for key, value in json_body.items()}
            request._body = json.dumps(
                sanitize_body, indent=2).encode('utf-8')
            request = await self.set_body(request)

        response = await call_next(request)
        return response

    async def set_body(self, request):
        receive_ = await request._receive()
        async def receive():
            return receive_

        request._receive = receive
        return request

    @staticmethod
    def sanitize_array(array_values):
        for index, value in enumerate(array_values):
            if isinstance(value, dict):
                array_values[index] = {key: bleach.clean(
                    value) if isinstance(value, str) else value for key, value in value.items()}
            else:
                array_values[index] = bleach.clean(value) if isinstance(value, str) else value
        
        return array_values



from arikaim.core.path import Path
from arikaim.core.utils import load_source
import os,importlib;
from sqlmodel import create_engine

class Db: 
    _instance = None

    def __init__(self):
        self._config = None
        self._engine = None

    def __new__(cls, *args, **kwargs):
        if not isinstance(cls._instance, cls):
            cls._instance = object.__new__(cls, *args, **kwargs)
        return cls._instance
    
    def connect(self, config):     
        self._config = config
        connect_link = 'mysql+pymysql://' + self._config['username'] + ':' + self._config['password'] + '@' + self._config['host'] + '/' +  self._config['database']      
        echo = False
        if echo in self._config:
            echo = self._config['echo']
        self._engine = create_engine(
            connect_link, 
            echo = echo, 
            pool_recycle = 3600,
            pool_pre_ping = True
        )
       
    def close(self):
        if not self._engine:
            return False
        self._engine.dispose()
        
        return True

    @property
    def engine(self):
        return self._engine
    
    def load_model(self,model_class, module_name, service_name = None):
        return load_model_class(model_class,module_name,service_name)


db = Db()




def load_model_class(model_class, module_name, service_name = None):
    if not service_name:        
        module = importlib.import_module('arikaim.core.db.models.' + module_name,model_class)
    else:
        path = os.path.join(Path.model(service_name),module_name)      
        module = load_source(module_name,path + '.py')

    return getattr(module,model_class)


def get_or_create(session, model, defaults = None, **kwargs):
    instance = session.query(model).filter_by(**kwargs).one_or_none()
    if instance:
        return instance, False
    else:
        kwargs |= defaults or {}
        instance = model(**kwargs)
        try:
            session.add(instance)
            session.commit()
        except Exception:  
            session.rollback()
            instance = session.query(model).filter_by(**kwargs).one()
            return instance, False
        else:
            return instance, True


from arikaim.core.collection.property import Property

class PropertiesDescriptor:

    def __init__(self):
        self._properties = {}
        self._result = {}
        self._title = ''
        self._description = ''

    @property
    def title(self):
        return self._title
    
    @title.setter
    def title(self, value: str):
        self._title = value

    @property
    def description(self):
        return self._description
    
    @description.setter
    def description(self, value: str):
        self._description = value

    def result(self, name: str):
        self._result[name] = Property(name)
        return self._result[name]
    
    @property
    def result_fields(self):
        return self._result

    def prop(self, name: str):
        self._properties[name] = Property(name)
        return self._properties[name]
    
    @property
    def props(self):
        return self._properties

    def param(self, name: str):
        self._properties[name] = Property(name)
        return self._properties[name]
    
    def get_example_values(self):
        result = {}
        for key, prop in self._properties.items():
            result[key] = prop.get_example_value()

        return result
    
    @property
    def params(self):
        return self._properties



import os
from datetime import datetime, timedelta, timezone
from redis import Redis
from rq import Queue as RqQueue
from rq import Worker

from arikaim.core.utils import load_source
from arikaim.core.logger import logger
from arikaim.core.queue.job import Job
from arikaim.core.path import Path

class Queue:

    def __init__(self):
        self._redis = Redis()
        self._queue = RqQueue(connection = self._redis)
        self._worker = None
       
    def boot(self):
        logger.info('Init qeueue')
     
    def run(self):
        logger.info('Run qeueue worker')
        self._worker = Worker(
            [self._queue], 
            connection = self._redis, 
            name = 'jobs'
        )
        self._worker.work()

    def add_job(self, job, *args, **kwargs):
        if isinstance(job,Job) == False: 
            raise Exception('Not valid job instance')
        
        if job.interval:
            return self._queue.enqueue_in(
                job.execute, 

                *args, 
                **kwargs
            )
        return self._queue.enqueue(job.execute, *args, **kwargs)
    
    def get_jobs_count(self):
        return len(self._queue)

    def remove_jobs(self):
        self._queue.empty()

    @property
    def jobs(self):
        return self._queue.jobs

    def create_job(self, job_class_name: str, module_name: str, service_name: str):
        path = os.path.join(Path.job(service_name),module_name)      
        module = load_source(module_name,path + '.py')

        job_class = getattr(module,job_class_name)
        if not job_class:
            return None
        
        return job_class()
    

queue = Queue()



import gc
import sys

from starlette.applications import Starlette
from starlette.exceptions import HTTPException
from starlette.middleware import Middleware

from arikaim.core.services import services
from arikaim.core.redis import redis_connect
from arikaim.core.utils import *
from arikaim.core.db.db import *
from arikaim.core.errors import error_handlers, http_exception, handle_error
from arikaim.core.logger import logger

from arikaim.core.admin.admin import AdminService
from arikaim.core.middleware.system import SystemMiddleware
from arikaim.core.middleware.exception_handler import ExceptionHandlerMiddleware

class ArikaimApp:
    _instance = None

    def __init__(self):             
        self._starlette = None
        self._config = None
        self._middlewares = []
        self._default_middlewares = [
            Middleware(SystemMiddleware),
            Middleware(ExceptionHandlerMiddleware)
        ]
    
    def __new__(cls, *args, **kwargs):
        if not isinstance(cls._instance, cls):
            cls._instance = object.__new__(cls, *args, **kwargs)
        return cls._instance
    
    def add_middleware(self, middleware):
        self._middlewares.append(middleware)

    def load_service_console_commands(self, service_name: str, module_name = None):
        if not module_name:
            module_name = 'console'

        console_file = os.path.join(Path.console(service_name),module_name)
        
        if os.path.isfile(console_file + '.py') == True: 
            # append service path  
            sys.path.append(Path.services(service_name))            
            return load_module('console',console_file + '.py')                              
        else:
            return False
        
    def load_console_commands(self, service_name, module_name = None):
        # load console module
        console = self.load_service_console_commands(service_name,'console')
        if console != False:
            call(console,'init')

        return self.load_service_console_commands(service_name, module_name)

    def load_config(self):
        self._config = load_module('config',os.path.join(Path.config(),'config.py'))
        if self._config is None:
            logger.error('Config file not found!')
            return False
        return self._config
    
    def system_init(self):   
        # include baase path 
        sys.path.insert(0,Path.base(relative = False))
        logger.info('Add base path ' + Path.base(relative = False))            
        # enable GC
        gc.enable()

        # load config
        if self.load_config() == False:
            return False

        # connect ro teds
        redis_connect(self._config.redis)

        # connect to db        
        db.connect(self._config.db)
       
    def boot(self):
        logger.info('Boot')

        self.system_init()
        # scan services
        services.scan_services()
      
        # load services routes
        services.boot_services()

        # load admin routes
        self.load_admin_routes()

        # middlewares
        middlewares = self._default_middlewares + self._config.middleware + self._middlewares
        for middleware in middlewares:
            logger.info('Add ' + str(middleware))
     
        # starlette instance
        self._starlette = Starlette(
            debug = True, 
            routes = services.routes, 
            on_shutdown = [self.on_shutdown],
            middleware = middlewares, 
            exception_handlers = error_handlers
        )
        self._starlette.add_exception_handler(HTTPException,http_exception)
        self._starlette.add_exception_handler(Exception,handle_error)
         
        return self.starlette

    async def on_shutdown(self):
        logger.info('Shutdown server')

        db.close()
        logger.info('Db connection closed')

    def load_admin_routes(self):
        logger.info('Load admin routes')

        admin = AdminService('admin','')
        admin.init_routes()

        services.routes.append(admin.routes)

    @property
    def starlette(self):
        return self._starlette

    @property
    def config(self):
        return self._config


app = ArikaimApp()



from arikaim.core.collection.property import Property

class Properties:

    def __init__(self, data):
        self._data = data
    
    def get_value(self, name: str, default = None):
        if name in self._data:
            return self._data[name]['value']
        else:
            return default
        
    @property
    def data(self):
        return self._data
    
    


import sys
from arikaim.core.logger import logger
from arikaim.core.utils import *
from arikaim.core.packages import load_package_descriptor
from arikaim.core.path import Path

class Services:
    _instance = None

    def __init__(self):        
        self._services = {}
        self._services_instance = {}
        self._routes = []
        
    def __new__(cls, *args, **kwargs):
        if not isinstance(cls._instance, cls):
            cls._instance = object.__new__(cls, *args, **kwargs)
        return cls._instance
    
    def __contains__(self, key: str) -> bool:
        return True if key in self._services.keys() else False

    def boot_services(self, load_routes: bool = True, init_container: bool = True):

        for service_name, service in self._services.items(): 
            if service['disabled'] == True:
                # skip disabled service boot
                logger.warning('service: ' + service_name + 'disabled ')
                continue

            logger.info('Boot service: ' + service_name)    
            # add to sys path
            sys.path.append(Path.services(service_name,False))      
            module_name = 'arikaim.services.' + service_name + '.' + service_name
    
            service_class = load_class(
                Path.services(service_name,relative = False),
                module_name,
                service_name.capitalize(),
                package = 'arikaim.services'
            )
            self._services_instance[service_name] = service_class(service_name) 
            self._services_instance[service_name].boot()
            
            if load_routes == True:
                self._services_instance[service_name].init_routes()
            if init_container == True:
                self._services_instance[service_name].init_container()
        
            # add service routes
            if self._services_instance[service_name].routes != None:
                self._routes.append(self._services_instance[service_name].routes)

    def scan_services(self):
        for file in os.scandir(Path.services()):
            if file.is_dir() and not file.name.startswith('.'):  
                package = load_package_descriptor(file.name)   
                package.setdefault('language','nodejs')
                package.setdefault('title',file.name)
                package.setdefault('disabled',False)

                if package['language'] != 'python':
                    # load only pyhton services
                    continue
               
                self._services[file.name] = { 
                    "name": file.name,   
                    "title": package['title'],   
                    "disabled": package['disabled']
                }

                # add service path to system paths
                sys.path.append(Path.services(file.name))

    def get_service(self, name: str):
        if name in self._services_instance:
            return self._services_instance[name]
        else:
            return None
        
    @property 
    def services(self):
        return self._services
    
    @property 
    def routes(self):
        return self._routes
    
services = Services()



from arikaim.core.controller import Controller, get
from arikaim.core.packages import load_package_descriptor

class ServiceDescriptor(Controller):

    @get
    async def get(self, request, data):  
        try:
            descriptor = load_package_descriptor(data['name'])
        except BaseException:
            self.error('Service not found')
            return
    
        self.field('name',data['name'])
        self.field('descriptor',descriptor)
    